# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''
共用类模块
'''

import binascii
import codecs
import importlib
import inspect
import io
import locale
import os
import pkg_resources
import re
import six
import sys
import threading
import time
import traceback

from xml.dom.minidom import Node
from datetime import datetime

from tuia.exceptions import TimeoutError

default_locale = locale.getdefaultlocale()
if default_locale:
    default_encoding = default_locale[1] or "utf-8"
else:
    default_encoding = "utf-8"
    
file_encoding = sys.getfilesystemencoding()
file_encoding_lower = file_encoding.lower()
for special_encoding in ["ansi", "ascii"]:
    if file_encoding_lower.find(special_encoding) >= 0:
        file_encoding = "utf-8"
        break
file_encoding = file_encoding or "utf-8"

class Timeout(object):
    '''TimeOut类，实现超时重试逻辑
    '''
    def __init__(self, timeout=10, interval=0.5):
        '''Constructor
        
        :param timeout: 超时秒数，默认是10
        :param interval: 重试时间间隔秒数，默认是0.5
        '''
        self.timeout=float(timeout)
        self.interval=float(interval)
        
    def retry(self,
        func,
        args,
        exceptions=(),
        resultmatcher=None,
        nothrow=False):
        """多次尝试调用函数，成功则并返回调用结果，超时则根据选项决定抛出TimeOutError异常。
        
        :param func: 尝试调用的函数
        :type args: dict或tuple
        :param args: func函数的参数
        :type exceptions: tuple类型，tuple元素是异常类定义，如QPathError, 而不是异常实例，如QPathError()
        :param exceptions: 调用func时抛出这些异常，则重试。
                                                                        如果是空列表(),则不捕获异常。
        :type resultmatcher: 函数指针类型 
        :param resultmatcher: 函数指针，用于验证第1个参数func的返回值。
                                                                                默认值为None，表示不验证func的返回值，直接返回。
                                                                               其函数原型为: 
                                  def result_match(ret):  # 参数ret为func的返回值
                                      pass                                                                                             
                                                                               当result_match返回True时，直接返回，否则继续retry。
        :type  nothrow:bool
        :param nothrow:如果为True，则不抛出TimeOutError异常
        
        :return: 返回成功调用func的结果
        :rtype : any
        """         
        start = time.time()
        waited = 0.0
        try_count = 0
        ret=None
        while True:
            try:
                try_count += 1
                if dict == type(args):
                    ret = func(**args)
                elif tuple == type(args):
                    ret = func(*args)
                else:
                    raise TypeError("args type %s is not a dict or tuple" % type(args))
                
                if resultmatcher == None or resultmatcher(ret)==True:
                    return ret
            except exceptions:
                pass
            
            waited = time.time() - start
            if waited < self.timeout:
                time.sleep(min(self.interval, self.timeout - waited))
            elif try_count ==1 :
                continue
            else:
                if nothrow:
                    return ret
                else:
                    raise TimeoutError("在%d秒里尝试了%d次" % (self.timeout, try_count))
    
    def waitObjectProperty(self, obj, property_name, waited_value, regularMatch=False):
        '''通过比较obj.property_name和waited_value，等待属性值出现。 
                             如果属性值obj.property_name是字符类型则waited_value做为正则表达式进行比较。
                             比较成功则返回，超时则抛出TimeoutError异常。
                             
        :param obj: 对象
        :param property_name: 要等待的obj对象的属性名
        :param waited_value: 要比较的的属性值，支持多层属性
        :param regularMatch: 参数 property_name和waited_value是否采用正则表达式的比较。
                                                                            默认为不采用（False）正则，而是采用恒等比较
        '''
        start = time.time()
        waited = 0.0
        try_count = 0
        isstr = isinstance(waited_value, six.string_types)
        while True:
            objtmp = obj    #增加多层属性支持
            pro_names = property_name.split('.')
            for i in range(len(pro_names)):
                propvalue = getattr(objtmp, pro_names[i])
                objtmp = propvalue
                
            if isstr and regularMatch:      #简化原逻辑
                if None != re.search(waited_value, propvalue):
                    return
            else:
                if waited_value == propvalue: 
                    return
            try_count +=1
            waited = time.time() - start
            if waited < self.timeout:
                time.sleep(min(self.interval, self.timeout - waited))
            else:
                raise TimeoutError("对象属性值比较超时（%d秒%d次）：期望值:%s，实际值:%s，" 
                                   % (self.timeout, try_count, waited_value, propvalue))

    def check(self, func, expect ):
        '''多次检查func的返回值是否符合expect设定的期望值，如果设定时间内满足，则返回True，否则返回False
        
        :param func: 尝试调用的函数
        :param expect: 设定的期望值
        
        :returns bool - 检查是否符合预期
        '''
        start = time.time()
        waited = 0.0
        while True:
            if expect == func():
                return True
            waited = time.time() - start
            if waited < self.timeout:
                time.sleep(min(self.interval, self.timeout - waited))
            else:
                return False
             
        
class Singleton(type):
    """单实例元类，用于某个类需要实现单例模式。
            使用方式示例如下::
          import six
          class MyClass(with_metaclass(Singleton, object)):
              def __init__(self, *args, **kwargs):
                  pass
    
    """
    _instances = {}
    def __init__(cls, name, bases, dic):
        super(Singleton, cls).__init__(name, bases, dic)
        cls._instances = {}
        
    def __call__(self, *args, **kwargs):
        if self not in self._instances:
            self._instances[self] = super(Singleton, self).__call__(*args, **kwargs)          
        return self._instances[self]
    
class LazyInit(object):
    '''实现延迟初始化
    
    使用方式示例::
    
        class _Win32Window(object)
            def click(self):
                #......
        class Control(object):
            def __init__(self, locator):
                self._locator = locator
                self._initobj = LazyInit(self, '_initobj', self._init_window)
            def _init_window(self):
                return _Win32Window(self._locator)
            def click(self):
                return self._initobj.click()
                
        ctrl = Control("/Name=xxx)
        ctrl.click()  # <-- call _init_window
        ctrl.click()
    
    '''
    def __init__(self, obj, propname, init_func):
        '''构造函数
        '''
        self.__obj = obj
        self.__propname = propname
        self.__init_func = init_func
        
    def __getattr__(self, attrname ):
        obj = self.__init_func()
        setattr(self.__obj, self.__propname, obj)
        attr = getattr(obj, attrname)
        del self.__obj
        del self.__propname
        del self.__init_func
        return attr
    
    def __setattr__(self, attrname, value):
        if attrname in ['_LazyInit__obj', '_LazyInit__propname', '_LazyInit__init_func']:
            return super(LazyInit, self).__setattr__(attrname, value)
        obj = self.__init_func()
        setattr(self.__obj, self.__propname, obj)
        setattr(obj, attrname, value)
        del self.__obj
        del self.__propname
        del self.__init_func
        
        
class ThreadGroupLocal(object):
    '''使用线程组本地存储的元类
    
    - 当配合ThreadGroupScope使用，类似threading.local()提供的TLS变种，一个线程和其子孙线程共享一个存储
    详细使用方式请参考ThreadGroupScope类
    
    - 当不在ThreadGroupScope中使用时，行为和threading.local()一致
    
    '''
    def __init__(self):
        curr_thread = threading.current_thread()
#         if not hasattr(curr_thread, 'qtaf_group'):
#             raise RuntimeError("current thread is not in any QTAF thread group scope")
#         self.__data = curr_thread.qtaf_local
        if hasattr(curr_thread, 'qtaf_group'):
            self.__data = curr_thread.qtaf_local
        else:
            if not hasattr(curr_thread, 'qtaf_local_outofscope'):
                curr_thread.qtaf_local_outofscope = {}
            self.__data = curr_thread.qtaf_local_outofscope
            
    def __setattr__(self, name, value ):
        if name.startswith('_ThreadGroupLocal__'):
            super(ThreadGroupLocal,self).__setattr__(name, value)
        else:
            self.__data[name] = value
        
    def __getattr__(self, name):
        if name.startswith('_ThreadGroupLocal__'):
            return super(ThreadGroupLocal,self).__getattr__(name)
        else:
            try:
                return self.__data[name]
            except KeyError:
                raise AttributeError("'ThreadGroupLocal' object has no attribute '%s'" % (name))

class ThreadGroupScope(object):
    '''指定线程组作用域，进入这个作用域的线程，以及在其作用域内创建的线程都同属于一个线程组
    
    使用示例如下::
    
        def _thread_proc():
            ThreadGroupLocal().counter +=1
            
        with ThreadGroupScope("test_group"):
            ThreadGroupLocal().counter = 0
            t = threading.Thread(target=_thread_proc)
            t.start()
            t.join()
            t = threading.Thread(target=_thread_proc)
            t.start()
            t.join()
            assert ThreadGroupLocal().counter == 2
        
    '''
    def __init__(self, name ):
        '''构造函数
        
        :param name: 线程组名称，全局唯一
        :type name: string
        '''      
        self._name = name
        
    def __enter__(self):
        curr_thread = threading.current_thread()
        if hasattr(curr_thread, 'qtaf_local'):
            raise RuntimeError("ThreadGroupScope cannot be nested")
        curr_thread.qtaf_local = {}
        curr_thread.qtaf_group = self._name
    
    def __exit__(self, *exc_info ):
        del threading.current_thread().qtaf_local
        del threading.current_thread().qtaf_group
    
    @staticmethod
    def current_scope():
        '''返回当前线程所在的线程组作用域，如果不存在于任务线程组作用域，则返回None
        '''
        curr_thread = threading.current_thread()
        if hasattr(curr_thread, 'qtaf_group'):
            return curr_thread.qtaf_group
 
_origin_thread_start_func = threading.Thread.start   

def _thread_start_func(self, *args, **kwargs):
    '''用于劫持threading.Thread.start函数
    '''
    curr_thread = threading.current_thread()
    if hasattr(curr_thread, 'qtaf_group'):
        self.qtaf_group = curr_thread.qtaf_group
        self.qtaf_local = curr_thread.qtaf_local
    return _origin_thread_start_func(self, *args, **kwargs)
    
threading.Thread.start = _thread_start_func



def ForbidOverloadMethods( func_name_list ):
    '''生成metaclass用于指定基类禁止子类重载函数
    '''
    class _metaclass(type):
        def __init__(cls, name, bases, dic):
            if len(bases) == 1 and bases[0] == object:
                super(_metaclass, cls).__init__(name, bases, dic)
            else:
                for it in func_name_list:
                    if it in dic.keys():
                        raise RuntimeError("不允许%s重载函数: %s" % (cls.__name__, it))
                super(_metaclass, cls).__init__(name, bases, dic)
                
    return _metaclass


class classproperty(object):
    '''类属性修饰器
    '''
    def __init__(self, getter ):
        self.getter = getter
    def __get__(self, instance, owner):
        return self.getter(owner)
    
def smart_text(s, decoding=None):
    '''convert any text or binary to text
    py2 text: utf-8 bytes
    py3 text: unicode
    '''
    if not isinstance(s, (six.string_types, six.binary_type)):
        raise RuntimeError("string or binary type didn't match with %r" % s)
    if six.PY3:
        if isinstance(s, six.text_type):
            return s
        else:
            try:
                return s.decode('utf8')
            except UnicodeDecodeError: # other encoding
                try:
                    if decoding is None:
                        decoding = default_encoding
                    return s.decode(decoding)
                except UnicodeDecodeError: # mixed encoding
                    return repr(s)
    else:
        return smart_binary(s, decoding=decoding) # py2
            
def smart_binary(s, encoding="utf8", decoding=None):
    '''convert any text or binary to binary of specified encoding 
    '''
    if not isinstance(s, (six.string_types, six.binary_type)):
        raise RuntimeError("string or binary type didn't match with %r" % s)
    if isinstance(s, six.text_type):
        return s.encode(encoding)
    
    assert type(s) == six.binary_type
    try:
        return s.decode("utf8").encode(encoding) # other encoding
    except UnicodeError:
        try:
            if decoding is None:
                decoding = default_encoding
            return s.decode(decoding).encode(encoding)
        except UnicodeError: # data mixed encoding
            if six.PY3:
                return bytes(repr(s), encoding)
            else:
                return bytes(repr(s))

def smart_bytes(data):
    if not isinstance(data, (six.string_types, six.binary_type)):
        raise ValueError("data=%r does not match string or bytes type" % data)
    if isinstance(data, six.text_type):
        return smart_binary(data)
    else:
        return data
            
def smart_to_hex(s, encoding="utf8", decoding=None):
    s = smart_binary(s, encoding=encoding, decoding=decoding)
    binary_s = binascii.hexlify(s)
    return smart_text(binary_s)

def smart_from_hex(s):
    s = smart_binary(s)
    binary_s = binascii.unhexlify(s)
    return smart_text(binary_s)

def smart_bytify(obj, encoding="utf-8", decoding=None):
    """recursively convert objects from string types to binary  
    """
    if isinstance(obj, dict):
        dic={}
        for key, value in obj.items():
            dic[smart_bytify(key, encoding, decoding)] = smart_bytify(value, encoding, decoding)
        return dic
    elif isinstance(obj, list):
        ls=[]
        for element in obj:
            ls.append(smart_bytify(element, encoding, decoding))
        return ls
    elif isinstance(obj, six.string_types):
        return smart_binary(obj, encoding, decoding)
    else:
        return obj
    
def smart_strfy(obj, decoding=None):
    """recursively convert objects from binary to text
    """
    if isinstance(obj, dict):
        dic={}
        for key, value in obj.items():
            dic[smart_strfy(key, decoding)] = smart_strfy(value, decoding)
        return dic
    elif isinstance(obj, list):
        ls=[]
        for element in obj:
            ls.append(smart_strfy(element, decoding))
        return ls
    elif isinstance(obj, six.string_types):
        return smart_text(obj, decoding)
    else:
        return obj

def get_thread_traceback(thread):
    '''获取用例线程的当前的堆栈
    
    :param thread: 要获取堆栈的线程
    :type thread: Thread
    '''
    for thread_id, stack in sys._current_frames().items():
        if thread_id != thread.ident:
            continue
        tb = "Traceback ( thread-%d possibly hold at ):\n" % thread_id
        for filename, lineno, name, line in traceback.extract_stack(stack):
            tb += '  File: "%s", line %d, in %s\n' % (filename, lineno, name)
            if line:
                tb += "    %s\n" % (line.strip())
        return tb
    else:
        raise RuntimeError("thread not found")
    
def get_method_defined_class(method):
    method_name = method.__name__
    need_prefix = False
    if method_name.startswith("__"):
        if method_name != "__init__":
            need_prefix = True
    if getattr(method, "__self__", None):    
        classes = [method.__self__.__class__]
    else:
        #unbound method
        if hasattr(method, "im_class"):
            classes = [method.im_class]
        else:
            classes = [getattr(method, "__class__", object)]
    while classes:
        c = classes.pop()
        if c == object:
            if len(classes) > 0:
                continue
        if need_prefix:
            target_name = "_" + c.__name__ + method_name
        else:
            target_name = method_name
        if target_name in c.__dict__:
            return c
        else:
            classes = list(c.__bases__) + classes
    raise RuntimeError("cannot find implement class for method: %s" % method)

def get_last_frame_stack(back_count=2):
    frame = inspect.currentframe()
    for _ in range(back_count):
        frame = frame.f_back
    stack = "".join(traceback.format_stack(frame, 1))
    return stack

def to_pretty_xml(doc, encoding="utf-8"):
    """we need to ensure each line to be binary type
    """
    class _XMLWriter(codecs.StreamWriter):
        """an inner writer to give writer a chance to handle each line
        """
        def write(self, data):
            data = smart_binary(data)
            self.stream.write(data)
    
    buff = io.BytesIO()
    indent = "    "
    newl = "\n"
    writer = _XMLWriter(buff)
    if doc.nodeType == Node.DOCUMENT_NODE: # document node needs encoding
        doc.writexml(writer, "", indent, newl, encoding)
    else:
        doc.writexml(writer, "", indent, newl)
    return writer.stream.getvalue()

def ensure_binary_stream(stream, encoding="utf-8"):
    if six.PY3:
        stream_type = type(stream)
        if not issubclass(stream_type, io.IOBase):
            raise ValueError("stream=%r does not match subclass of io.IOBase" % stream)
        if not hasattr(stream, "mode"):
            raise ValueError("stream=%r does not have attribute mode" % stream)
        if "b" in stream.mode:
            new_stream = stream
        else:
            new_stream = stream.buffer
    else:
        if getattr(stream, "encoding", None):
            if not stream.encoding.lower().startswith('ansi'): # linux ascii
                encoding = stream.encoding
        new_stream = stream
    return new_stream, encoding

def codecs_open(filename, mode="rb", encoding=None, errors="strict", buffering=1):
    filename = smart_binary(filename, encoding=file_encoding)
    return codecs.open(filename, mode=mode, encoding=encoding, errors=errors, buffering=buffering)

def path_exists(filename):
    filename = smart_binary(filename, encoding=file_encoding)
    return os.path.exists(filename)

def get_os_version():
    if sys.platform == 'win32':
        with os.popen("ver") as pipe:
            osver = smart_text(pipe.read())
    else:
        osver = smart_text(str(os.uname()))  # @UndefinedVariable
    return osver

if six.PY3:
    maketrans_func = str.maketrans
else:
    import string
    maketrans_func = string.maketrans
        
BAD_FILE_CHARS = r'\/*?:<>"|~'
BAD_VAR_CHAR = BAD_FILE_CHARS + '()[]+-=& '
TRANS = maketrans_func(BAD_VAR_CHAR, '_'*len(BAD_VAR_CHAR))
BAD_VAR_CHAR_SET = set(BAD_VAR_CHAR)

def translate_bad_char(input_string):
    if six.PY2:
        translated_string = smart_binary(input_string).translate(TRANS)
    else:
        translated_string = smart_text(input_string).translate(TRANS)
    return translated_string

def has_bad_char(input_string):
    if set(input_string) & BAD_VAR_CHAR_SET:
        return True
    else:
        return False

def get_time_str():
    now = datetime.now()
    time_str = now.strftime("%Y%m%d_%H%M%S%f")[:-3]
    return time_str

def get_inner_resource(resource_module, resource_name):
    if not os.path.isfile(__file__): # using egg
        file_path = pkg_resources.resource_filename(resource_module, resource_name)
    else:
        mod = importlib.import_module(resource_module)
        mod_path = mod.__path__[0]
        file_path = os.path.join(mod_path, resource_name)
    if not path_exists(file_path):
        raise RuntimeError("resource path: %s not existed." % file_path)
    return file_path

def get_attribute_from_string(object_path):
    parts = object_path.split(".")
    parts_len = len(parts)
    mod = None
    for index in range(parts_len):
        mod_path = ".".join(parts[:index + 1])
        try:
            mod = importlib.import_module(mod_path)
        except ImportError:
            break
    value = mod
    for new_index in range(index, parts_len):
        try:
            value = getattr(value, parts[new_index])
        except AttributeError:
            raise AttributeError("%s has no attribute or submodule named \"%s\"" % (value, parts[new_index]))
    return value
        
if __name__ == "__main__":
    pass


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''log模块
'''

import logging
import sys
import traceback
from testbase import context
from testbase.util import ensure_binary_stream, smart_binary
    
_stream, _encoding = ensure_binary_stream(sys.stdout)

class _Formatter(logging.Formatter):
    def format(self, record):
        s = super(_Formatter, self).format(record)
        return smart_binary(s, encoding=_encoding)
    
_stream_handler=logging.StreamHandler(_stream)
_stream_handler.terminator = b"\n"
_stream_handler.setFormatter(_Formatter())

class TestResultBridge(logging.Handler):
    '''中转log信息到TestResult
    '''
    def emit(self, log_record):
        '''Log Handle 必须实现此函数
        '''
        testresult = context.current_testresult()
        if testresult is None:
            _stream_handler.emit(log_record)
            return
        record = {}
        if log_record.exc_info:
            record['traceback'] = ''.join(traceback.format_tb(log_record.exc_info[2])) + '%s: %s' %(
                                   log_record.exc_info[0].__name__, log_record.exc_info[1])
        testresult.log_record(log_record.levelno, log_record.msg, record)
        
_LOGGER_NAME = "QTA_LOGGER"
_logger = logging.getLogger(_LOGGER_NAME)
_logger.setLevel(logging.DEBUG)
_logger.addHandler(TestResultBridge())

        
def critical(msg, *args, **kwargs):
    _logger.error(msg, *args, **kwargs)
 
fatal = critical

def error(msg, *args, **kwargs):
    '''Log a message with severity 'ERROR' on the root logger.
    '''
    _logger.error(msg, *args, **kwargs)

def exception(msg, *args):
    '''Log a message with severity 'ERROR' on the root logger,with exception information.
    '''
    _logger.exception(msg, *args)

def warning(msg, *args, **kwargs):
    '''Log a message with severity 'WARNING' on the root logger.
    '''
    _logger.warning(msg, *args, **kwargs)

warn = warning

def info(msg, *args, **kwargs):
    '''Log a message with severity 'INFO' on the root logger.
    '''
    _logger.info(msg, *args, **kwargs)

def debug(msg, *args, **kwargs):
    '''Log a message with severity 'DEBUG' on the root logger.
    '''
    _logger.debug(msg, *args, **kwargs)

def log(level, msg, *args, **kwargs):
    '''Log 'msg % args' with the integer severity 'level' on the root logger.
    '''
    _logger.log(level, msg, *args, **kwargs)
    
def addHandler(hdlr):
    '''Add the specified handler to this logger.
    '''
    _logger.addHandler(hdlr)
    
def removeHandler(hdlr):
    '''Remove the specified handler from this logger.
    '''
    _logger.removeHandler(hdlr)



#-*- coding: utf-8 -*
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

import warnings
warnings.warn("`tuia.remoteprocessing` is depressed, please use `qt4c.remoteprocessing` instead", DeprecationWarning)
from qt4c.remoteprocessing import *


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''
扩展库管理（仅独立模式使用）
'''

import os
import zipfile
import shutil
import logging
import xml.dom.minidom as dom

from testbase.util import codecs_open

class ExLibManager(object):
    '''扩展库管理器
    '''
    def __init__(self, proj_root ):
        self._top_dir = proj_root
        self._installed_libs = os.path.join(self._top_dir, 'exlib', 'installed_libs.txt')
    
    def _get_egg_name(self, egg_path ):
        '''获取egg包名称
        '''
        with zipfile.ZipFile(egg_path, 'r') as egg:
            with egg.open('EGG-INFO/PKG-INFO', 'r') as fd:
                for line in fd.readlines():
                    if line.startswith('Name:'):
                        return line.split(':')[1].strip()

    def install(self, egg_path ):
        '''安装
        '''
        #获取全部的顶级package
        toplv_pkgs = []
        with zipfile.ZipFile(egg_path, 'r') as egg:
            for f in egg.filelist:
                items = f.filename.split('/')
                if len(items) == 2 and items[1].lower() == '__init__.py':
                    toplv_pkgs.append(items[0])
        
        exlib_dir = os.path.join(self._top_dir, 'exlib')
        if not os.path.isdir(exlib_dir):
            os.makedirs(exlib_dir)
        
        #移除老的包
        egg_name = self._get_egg_name(egg_path)
        egg_paths_remove = []
        if egg_name:
            for file_name in os.listdir(exlib_dir):
                if not file_name.lower().endswith('.egg'):
                    continue
                file_path = os.path.join(exlib_dir, file_name)
                if not os.path.isfile(file_path):
                    continue
                if egg_name == self._get_egg_name(file_path):
                    os.remove(file_path)
                    egg_paths_remove.append(file_path)
        
        #拷贝包
        shutil.copy(egg_path, exlib_dir)
        
        #记录到installed_libs.txt
        installed_pkgs = self.list_names()
        for pkg in toplv_pkgs[:]:
            if pkg in installed_pkgs:
                toplv_pkgs.remove(pkg)
        if toplv_pkgs:
            with codecs_open(self._installed_libs, 'a+') as fd:
                for pkg in toplv_pkgs:
                    fd.write(pkg+'\n')
                    
        self._update_pydev_conf(egg_path, egg_paths_remove)
                    
    def _update_pydev_conf(self, egg_path, egg_paths_remove):
        '''修改pydev配置
        '''
        egg_names_remove = []
        for it in egg_paths_remove:
            egg_names_remove.append(os.path.basename(it).lower())
        egg_name = os.path.basename(egg_path)
        
#         if len(egg_names_remove) == 1 and egg_names_remove[0] == egg_name:
#             return
        
        pydev_path = os.path.join(self._top_dir, '.pydevproject')
        if not os.path.isfile(pydev_path):
            logging.warn('pydev configure file not found')
            return
        doc = dom.parse(pydev_path)
        for propnode in doc.getElementsByTagName('pydev_pathproperty'):
            if propnode.getAttribute('name') == 'org.python.pydev.PROJECT_SOURCE_PATH':
                break
        else:
            propnode = doc.createElement('pydev_pathproperty')
            propnode.setAttribute('name', 'org.python.pydev.PROJECT_SOURCE_PATH')
            projnodes = doc.getElementsByTagName('pydev_project')
            if not projnodes:
                logging.warn('pydev configure file corrupted')
                return
            projnodes[0].AppendChild(propnode)
        

        for pathnode in propnode.getElementsByTagName('path'):
            file_path =  pathnode.childNodes[0].data
            name = file_path[file_path.rfind('/')+1:]
            if name.lower() in egg_names_remove:
                propnode.removeChild(pathnode)
                
        pathnode = doc.createElement('path')
        print(egg_name)
        pathnode.appendChild(doc.createTextNode('/${PROJECT_DIR_NAME}/exlib/%s'%egg_name))
        propnode.appendChild(pathnode)
        with codecs_open(pydev_path, 'w', 'utf8') as fd:
            fd.write(doc.toxml(encoding='UTF-8'))
        
                    
    def list_names(self):
        '''获取全部的扩展包的名字
        '''
        if not os.path.isfile(self._installed_libs):
            return []
        names = []
        with codecs_open(self._installed_libs, 'r', encoding="utf-8") as fd:
            for libname in fd.readlines():
                libname = libname.strip('\r\n ')
                if not libname:
                    continue
                names.append(libname)
        return names


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''
QPath解析器

QPath是一个用于定位各个平台的UI控件（除Web控件）的查询语言。

 - QPath语法定义如下::

    QPath ::= Seperator QPath Seperator UIObjectLocator
    UIObjectLocator ::= UIObjectProperty PropertyConnector UIObjectLocator
    UIObjectProperty ::= UIProperty 
                        | RelationProperty 
                        | IndexProperty
                        | UITypeProperty 
    UIProperty ::= PropertyName Operator Literal
    RelationProperty ::= MaxDepthIdentifier EqualOperator Literal
    IndexProperty ::= InstanceIdentifier EqualOperator Literal
    UITypeProperty ::= UITypeIdentifier EqualOperator StringLiteral
    MaxDepthIdentifier := "MaxDepth"
    InstanceIdentifier ::= "Instance"
    UITypeIdentifier := "UIType"
    Operator ::= EqualOperator | MatchOperator
    PropertyConnector ::= "&&"
    Seperator ::= "/"
    EqualOperator ::= "="  #精确匹配
    MatchOperator ::= "~="  #使用正则进行模糊匹配
    PropertyName  ::= "[a-zA-Z_]*"
    Literal := StringLiteral
              | IntegerLiteral
              | BooleanLiteral
    StringLiteral ::= "\"[a-zA-Z_]*\""
    IntegerLiteral ::= "[0-9]*"
    BooleanLiteral := "True" | "False" | "true" | "false"


 - 需要注意的是，QPath的属性名都是大小写无关的。

 - 简单举例如下::
 
    / ClassName='TxGuiFoundation' && Caption~='QQ\d+' / name='mainpanel'


'''
from __future__ import absolute_import

import six
import types

try:
    from .ply import lex, yacc
    from .ply.lex import TOKEN
except ImportError:
    from ply import lex, yacc
    from ply.lex import TOKEN


class QPathSyntaxError(Exception):
    '''QPath语法错误
    '''

    def __init__(self, qpath_string, err_msg, lexpos):
        '''Constructor
        
        :param qpath_string: QPath字符串
        :type qpath_string: str
        :parma err_msg: 错误信息
        :type err_msg: str
        :param lexpos: 错误对应的词法位置
        :type lexpos: int
        '''
        self.qpath_string = qpath_string
        self.msg = err_msg
        self.lexpos = lexpos

    def __str__(self):
        return '%s\n  %s\n  %s^' % (self.msg, self.qpath_string, ' ' * self.lexpos)


class QPathLexer(object):
    '''QPath词法解析器
    '''

    tokens = ['SEPERATOR',
              'EQUAL',
              'MATCH',
              'AND',
              'BOOL_CONST',
              'STRING_LITERAL',
              'INT_CONST_DEC',
              'INT_CONST_OCT',
              'INT_CONST_HEX',
              'PROPERTY',
              'MINUS']

    t_ignore = "\t "
    t_EQUAL = r'='
    t_MATCH = r'~='
    t_AND = r'(&&)|(&)'
    t_MINUS = r'-'
    t_SEPERATOR = '/'

    bad_match = r'~[^=*]'

    t_PROPERTY = r'[a-zA-Z_][0-9a-zA-Z_]*'

    boolean_constant = '(True)|(False)|(true)|(false)'

    decimal_constant = '(0)|([1-9][0-9]*)'
    octal_constant = '0[0-7]+'
    hex_prefix = '0[xX]'
    hex_digits = '[0-9a-fA-F]+'
    hex_constant = hex_prefix + hex_digits

    bad_octal_constant = '0[0-7]*[89]'

    # simple_escape = r"""([a-zA-Z._~!=&\^\-\\?'"])"""
    # decimal_escape = r"""(\d+)"""
    # hex_escape = r"""(x[0-9a-fA-F]+)"""
    # bad_escape = r"""([\\][^a-zA-Z._~^!=&\^\-\\?'"x0-7])"""
    # escape_sequence = r"""(\\("""+simple_escape+'|'+decimal_escape+'|'+hex_escape+'))'

    escape_sequence_1 = r'''(?<=\\)"'''
    escape_sequence_2 = r"""(?<=\\)'"""

    string_char_1 = r"""([^"]|""" + escape_sequence_1 + ')'
    string_literal_1 = '"' + string_char_1 + '*"'
    # bad_string_literal_1 = '"'+string_char_1+'*'+bad_escape+string_char_1+'*"'

    string_char_2 = r"""([^']|""" + escape_sequence_2 + ')'
    string_literal_2 = "'" + string_char_2 + "*'"
    # bad_string_literal_2 = "'"+string_char_2+'*'+bad_escape+string_char_2+"*'"

    string_literal = '(' + string_literal_1 + ')|(' + string_literal_2 + ')'
    # bad_string_literal = '('+bad_string_literal_1+')|('+bad_string_literal_2+')'

    @TOKEN(boolean_constant)
    def t_BOOL_CONST(self, t):
        if t.value.lower() == 'true':
            t.value = True
        else:
            t.value = False
        return t

    @TOKEN(hex_constant)
    def t_INT_CONST_HEX(self, t):
        t.value = int(t.value, 16)
        return t

    @TOKEN(octal_constant)
    def t_INT_CONST_OCT(self, t):
        t.value = int(t.value, 8)
        return t

    @TOKEN(decimal_constant)
    def t_INT_CONST_DEC(self, t):
        t.value = int(t.value)
        return t

    @TOKEN(string_literal)
    def t_STRING_LITERAL(self, t):
        col = t.value[0]
        value = t.value[1:-1].replace("\\%s" % col, col)
        t.value = value
        return t

#     @TOKEN(bad_string_literal)
#     def t_BAD_STRING_LITERAL(self, t):
#         self._error("字符串包含非法的转移字符", t)

    @TOKEN(bad_match)
    def t_BAD_MATCH(self, t):
        self._error("'~'后只能连接'='", t)

    def t_error(self, t):
        msg = '存在非法字符: %s' % repr(t.value[0])
        self._error(msg, t)

    def _error(self, msg, t):
        '''错误通知
        '''
        raise QPathSyntaxError(self._qpath_string, msg, t.lexpos)

    def input(self, qpath_string):  # @ReservedAssignment
        '''词法分析输入
        
        :param qpath_string: QPath字符串
        :type qpath_string: str
        '''
        self._qpath_string = qpath_string
        self._lexer = lex.lex(object=self)
        self._lexer.input(qpath_string)
        return self

    def token(self):
        '''解析并返回一个Token
        '''
        t = self._lexer.token()
        return t


class PropertyName(object):
    '''QPath属性名
    '''

    def __init__(self, value, lexpos):
        '''Constructor
        
        :param value: 属性名字符串
        :type value: str
        :param lexpos: 对应的词法位置
        :type lexpos: int
        '''
        self.value = value
        self.lexpos = lexpos

    def __str__(self):
        return '<PropertyName value:%s lexpos:%s>' % (self.value, self.lexpos)


class Literal(object):
    '''QPath属性常量值
    '''

    def __init__(self, value, lexpos):
        '''Constructor
        
        :param value: 属性值常量
        :type value: str/int/bool
        :param lexpos: 对应的词法位置
        :type lexpos: int
        '''
        self.value = value
        self.lexpos = lexpos

    def __str__(self):
        return '<Literal value:%s lexpos:%s>' % (repr(self.value), self.lexpos)


class Operator(object):
    '''QPath操作符
    '''

    def __init__(self, value, lexpos):
        '''Constructor
        
        :param value: 属性操作符
        :type value: str
        :param lexpos: 对应的词法位置
        :type lexpos: int
        '''
        self.value = value
        self.lexpos = lexpos

    def __str__(self):
        return '<Operator value:%s lexpos:%s>' % (repr(self.value), self.lexpos)


class UIObjectProperty(object):
    '''QPath属性
    '''

    def __init__(self, name, operator, value):
        '''Constructor
        
        :param name: 属性名
        :type name: PropertName
        :param operator: 属性操作符
        :type operator: Operator
        :param value: 属性值
        :type value: Literal
        '''
        self.name = name
        self.operator = operator
        self.value = value
        self.lexpos = self.name.lexpos

    def __str__(self):
        return "<UIObjectProperty %s>" % (self.format())

    def format(self):
        '''格式化字符串
        
        :returns: str
        '''
        return '%s%s%s' % (self.name.value, self.operator.value, repr(self.value.value))


class UIObjectLocator(object):
    '''QPath Locator
    '''

    def __init__(self, properties):
        '''Constructor
        
        :param properties: 属性字典
        :type properties: dict
        '''
        self._prop_dict = {}
        for it in properties:
            self._prop_dict[it.name.value.upper()] = it
        self.lexpos = properties[0].lexpos

    def append(self, prop):
        '''增加一个属性
        
        :param prop: 属性
        :type prop: UIObjectProperty
        '''
        self._prop_dict[prop.name.value.upper()] = prop

    def dumps(self):
        '''序列化
        
        :returns: list
        '''
        d = {}
        for name in self:
            prop = self[name]
            d[prop.name.value] = [prop.operator.value, prop.value.value]
        return d

    def format(self):
        '''格式化字符串
        
        :returns: str
        '''
        return ' & '.join([it.format() for it in self._prop_dict.values()])

    def __str__(self):
        return '<UIObjectLocator "%s">' % self.format()

    def __getitem__(self, name):
        return self._prop_dict.get(name.upper())

    def __delitem__(self, name):
        del self._prop_dict[name.upper()]

    def __setitem__(self, name, val):
        self._prop_dict[name.upper()] = val

    def __contains__(self, name):
        return name.upper() in self._prop_dict

    def __iter__(self):
        return self._prop_dict.__iter__()


class QPathParser(object):
    '''QPath语法解析器
    
    QPath解析器解析QPath后会生成两个结构:
    
    1. QPath结构列表：每一个UIObjectLocator用一个字典表示
    其中字典的键值为属性名，对应的值为一个长度为2的列表，第一个元素为操作符号，
    目前为字符串'='或'~='，第二个元素为属性值
    
    2. QPath词法位置信息：和解析后的结构对应，每一个UIObjectLocator用一个字典表示
    其中字典的键值为属性名，对应的指为一个长度为3的列表，第一个元素为属性名的
    词法位置，第二个元素为操作符的词法位置，第三个元素为属性值的词法位置
    
    比如以下的QPath::
    
        / ClassName="TxGuiFoundation" && Caption1~='QQ\d+' && Instance=-1 / UIType='GF' && name='mainpanel' && MaxDepth=10
    
    解析后得到的结构列表为::
    
        [{'Caption1': ['~=', 'QQ\\d+'],
          'ClassName': ['=', 'TxGuiFoundation'],
          'Instance': ['=', -1]},
         {'MaxDepth': ['=', 10], 
          'UIType': ['=', 'GF'], 
          'name': ['=', 'mainpanel']}]

    解析后得到的词法位置信息表为::
    
        [{'CAPTION1': [33, 41, 43],
          'CLASSNAME': [2, 11, 12],
          'INSTANCE': [54, 62, 63]},
         {'MAXDEPTH': [103, 111, 112], 
          'NAME': [83, 87, 88], 
          'UITYPE': [68, 74, 75]}]
    
    使用方法示例::
    
        qp ="""/ ClassName="TxGuiFoundation" && Caption1~='QQ\d+' && Instance='-1' / UIType='GF' && name='mainpanel' && MaxDepth='10'"""
        parser = QPathParser()
        qpath_struct, lex_info = parser.parse(qp)
    '''

    INT_TYPE_PROPNAMES = ['INSTANCE', 'MAXDEPTH']

    class _NullStream(object):
        '''模拟空设备
        '''

        def write(self, *_):
            pass

    def __init__(self, verbose=False):
        '''构造函数
        
        :param verbose: 是否有log提示
        :type verbose: boolean
        '''
        if verbose:
            self._logger = None
        else:
            self._logger = yacc.PlyLogger(self._NullStream())

    def parse(self, qpath_string):
        '''返回解析后的结果
        
        :param qpath_string: QPath字符串
        :type qpath_string: string
        :returns: list, list - 解析后结构, 词法位置信息
        '''
        self._last_locator = None
        qpath_string = qpath_string.strip()
        self._lexer = QPathLexer()
        self.tokens = self._lexer.tokens
        self._parser = yacc.yacc(module=self, debuglog=self._logger, errorlog=self._logger, write_tables=0)
        self._qpath_string = qpath_string
        parsed_structs = []
        lex_structs = []
        for locator in self._parser.parse(qpath_string, self._lexer):
            parsed_structs.append(locator.dumps())
            lex_struct = {}
            for propname in locator:
                prop = locator[propname]
                lex_struct[prop.name.value] = [prop.name.lexpos, prop.operator.lexpos, prop.value.lexpos]
            lex_structs.append(lex_struct)
        return parsed_structs, lex_structs

    def _error(self, msg, p, pos):
        '''提示错误
        '''
        raise QPathSyntaxError(self._qpath_string, msg, pos)

    # 以下为YACC语法构造函数
    def p_qpath(self, p):
        '''qpath : SEPERATOR qpath_content
        '''
        p[0] = p[2]

    def p_qpath_content(self, p):
        '''qpath_content : 
                         | object_locator
                         | qpath_content SEPERATOR object_locator
        '''
        if len(p) == 2:
            p[0] = [p[1]]
        else:
            p[1].append(p[3])
            p[0] = p[1]

    def p_object_locator(self, p):
        '''object_locator : prop
                          | object_locator AND prop
        '''
        if len(p) == 2:
            p[0] = UIObjectLocator([p[1]])
        else:
            p[0] = p[1]
            p[0].append(p[3])

    def p_prop(self, p):
        '''prop : prop_name operator prop_value
        '''
        if p[1].value.upper() in self.INT_TYPE_PROPNAMES:
            if p[2].value == '~=':
                self._error('"%s"属性不可以使用"~="操作符' % (p[1].value), p[2], p[2].lexpos)
            if not isinstance(p[3].value, int):
                try:
                    p[3].value = int(p[3].value)
                except ValueError:
                    self._error('"%s"属性值不可为"%s"类型，必须为int类型' % (p[1].value, type(p[3].value)), p[3], p[3].lexpos)
            if p[1].value.upper() == 'MAXDEPTH':
                if p[3].value <= 0:
                    self._error("MaxDepth属性值必须>0", p[3], p[3].lexpos)

        elif p[2].value == '~=':
            if not isinstance(p[3].value, six.string_types):
                self._error('操作符"~="不可以连接"%s"类型的属性' % (type(p[3].value)), p[2], p[2].lexpos)

        p[0] = UIObjectProperty(p[1], p[2], p[3])

    def p_prop_name(self, p):
        '''prop_name : PROPERTY
        '''
        p[0] = PropertyName(p[1], p.lexpos(1))

    def p_operator(self, p):
        '''operator : EQUAL
                    | MATCH
        '''
        p[0] = Operator(p[1], p.lexpos(1))

    def p_prop_value(self, p):
        '''prop_value : STRING_LITERAL
                      | BOOL_CONST
                      | int_const
        '''
#         if isinstance(p[1], types.StringTypes) and len(p[1]) == 0:
#             self._error("属性值不可为空", p, p.lexpos(1))
        if isinstance(p[1], Literal):
            p[0] = p[1]
        else:
            p[0] = Literal(p[1], p.lexpos(1))

    def p_int_const(self, p):
        '''int_const : INT_CONST_DEC
                     | INT_CONST_OCT
                     | INT_CONST_HEX
                     | MINUS int_const
        '''
        if len(p) == 2:
            p[0] = Literal(p[1], p.lexpos(1))
        else:
            p[0] = Literal(-p[2].value, p.lexpos(1))

    def p_error(self, p):
        '''处理错误
        '''
        if p is None:
            self._error("不完整的QPath", p, len(self._qpath_string))
        lexpos = p.lexpos
        if not isinstance(lexpos, types.IntType):  # 不是Token
            lexpos = p.lexpos(1)
        self._error("QPath语法错误", p, lexpos)


if __name__ == '__main__':

    q = QPathParser().parse("/classname='UIATableCell' && label~='XXX.*群，\\d，ZZZ.*\:.+，.*\\d\:\\d' && visible=true")
    x = "/classname='UIATableCell' && label~='XXX.*群，\\d.*\\,' && visible=true"

    QPathParser().parse('/class="\\d\\""')

    QPathParser().parse("/class='\\d\:\"\\'xxx'")
